# This file performs the analysis of the refugee paper

using LazyArrays
using Plots
using StatsPlots

#include("Codes_restructured.jl")
include("refugees_codes.jl")

# Load Parameters
println("Load Parameters")
paramDF = CSV.read("refugees_parameters.csv",DataFrame);
paramDF

pm_estimates = zeros(1,length(paramDF[1,:]))
for j = 1:length(pm_estimates)
pm_estimates[j] = paramDF[1,j]
end

# Load Data
println("Load Data")
Data_in=construct_data("refugees_data.csv");
Len=size(Data_in.DF,1);

# Moments from data
True_Moments=(Ref_Wage_Ratio_log=-0.075,
            Ref_manu_gr_50=0.317,
            Ref_manu_gr_61=0.241,
            Ref_GDP_gr_50=-0.083,
            Ref_GDP_gr_61=0.502,
            Ref_pop_gr_50=1.359,
            Ref_pop_gr_61=1.029,
            Ref_sh_cor_50=0.735,
            Ref_sh_cor_61=0.586,
            GDP_lag_coef=0.555,
            GDP_lag_sd=0.041,
            Manuf_Ag_WageRatio=1.5,
            In_Out_Flow_Ratio=0.73,
            Mobility_Distance=0.67,
            Distance_coef=0.06,
            Ref_pop_gr_50_55=-0.342,
            Ref_pop_gr_55_61=-0.087,
            Ref_GDP_gr_80=0.201,
            pop_coef_80_UB=1.2,
            Ref_pop_gr_80=1.0
            )


# Calibrate fundamentals and get inputs for model
println("Calibrate Fundamentals")
eq_inputs=fundamentals_restr(pm_estimates,ones(Len),ones(Len),Data_in.DF,Data_in.Distance_Matrix);

# Get productivity series 
#NumPeriods = 80; NumSim = 50;
NumPeriods = 80; NumSim = 3;
Q_input=Qcube_restr(eq_inputs.pm,eq_inputs.Q,NumPeriods,NumSim;shocks=true);

# Calculate equilibrium
println("Calculate Equilibrium path and Moments")
eq_output=equilibrium_allocation(eq_inputs.pm,eq_inputs.T,eq_inputs.exogV,
    Q_input,eq_inputs.N,eq_inputs.L_I,eq_inputs.L_F,
    Data_in.DF[1:end-1,:refugee_share_1950],
    eq_inputs.ag_wages,eq_inputs.manuf_wages,eq_inputs.TradeCosts,
    eq_inputs.MobilityCosts,eq_inputs.Ψ; convergence_psi=0.8);



####################################################################
####### Generate output from model
####################################################################

####### Paper

# Extract parameters and moments (Table 11)
println("Compute parameters and moments (Table 11)")
export_params(eq_inputs.pm, "ModelOutput/Table11_parameters.csv")
moments_estimates = compute_moments(eq_inputs.pm,Data_in.DF,eq_output);
export_moments(True_Moments,moments_estimates, "ModelOutput/Table11_moments.csv" )

# Calculate scale effect coefficients (Figure 4)
println("Estimate scale effects for income and population growth (Figure 4)")
scale_effects(eq_inputs.pm,Data_in.DF,eq_output,"ModelOutput/Figure4_scale_coefficients.csv");

# Aggregate and spatial implications (Figure 5)
println("Compute aggregate and spatial implications (Figure 5)")
#NumSimBSAggregate = 100;
NumSimBSAggregate = 3;

agg_implications(eq_inputs,"ModelOutput/Figure5_aggregate_implications.csv",Data_in.DF;sims=NumSimBSAggregate,end_yr=2000);
spatial_implications(eq_inputs,"ModelOutput/Figure5_spatial_implications.csv",Data_in.DF;sims=1);

# Export data from model to study spatial sorting (Table 12)
println("Export data from model to study spatial sorting (Table 12)")
make_analysis_data(eq_inputs.pm,Data_in.DF,eq_output,"ModelOutput/Table12_model_data.csv")

# Supply and Demand (Figure 6)
println("Compute supply and demand decomposition (Figure 6)")
#supply_demand(eq_inputs,"ModelOutput/Figure6_supply_demand.csv",Data_in.DF)

# Rural industrialization (Figure 7)
println("Compute persistence of rural industrialization (Figure 7)")
allocation_cf(eq_inputs,"ModelOutput/Figure7_rural_industrialization", Data_in.DF; end_yr = 2000);


####### Appendix

# Identification of parameters (Figures SM6 & SM7)
println("Construct data for plots for identification of parameters (Figures SM 6 & SM 7)")
identification(pm_estimates,Data_in, "ModelOutput/FigureSM6_7_id")

# Stationarity of the model (Figures SM8 & SM9)
println("Compute long-run implications in baseline model (Figures SM8 & SM9)")
#NumSimBSLongRun = 30;
NumSimBSLongRun = 4;

long_run_implications(eq_inputs,Data_in.DF,"baseline","ModelOutput/Figure";sims=NumSimBSLongRun,end_yr=2200)

# Compare with model with more potent convergence (Figure SM9)
println("Compute long-run implications in model without path dependence (Figure SM9)")
pm_convergence = copy(pm_estimates);
pm_convergence[5] = 1;      #Low epsilon
pm_convergence[13] = 0.6;   #High congestion in amenities

eq_inputs_convergence=fundamentals_restr(pm_convergence,ones(Len),ones(Len),Data_in.DF,Data_in.Distance_Matrix);
long_run_implications(eq_inputs_convergence,Data_in.DF,"convergence","ModelOutput/Figure";sims=NumSimBSLongRun,end_yr=2010)
