
using Random
using SpecialFunctions
using DataFrames
using Optim
using Parameters
using FixedEffectModels
using Pkg
using LinearAlgebra
using Parameters
using Statistics
using StatsBase
using CSV


# store parameters and it add default values
@with_kw struct param
    θ::Float64      = 2.0
    χ::Float64      = .6
    ϕ_fa::Float64   = 1.0
    ϕ_fm::Float64   = 1.0
    ϕ_ia::Float64   = .7
    ϕ_im::Float64   = 37
    ψ::Float64      = .04
    ϵ::Float64      = 3.6
    σ::Float64      = 5.0
    α::Float64      = .2
    ɷ::Float64      = 0.0
    ϱ::Float64      = 0.0
    ρ::Float64      = 2.63
    γ::Float64      = .5
    β::Float64      = .6
    λ::Float64      = .7
    δ::Float64      = .2
    η::Float64      = .95
    ρ_a::Float64    = .1
    fE::Float64     = 1.0
end


# stores output of the calibration code
mutable struct Calibration_Output{T}
    bwguess_F::Vector{T}
    bwguess_I::Vector{T}
    iter1::Int
    tol1::Float64
    V::Vector{T}
    pi_ia::Vector{T}
    pi_fa::Vector{T}
    pi_im::Vector{T}
    pi_fm::Vector{T}
    ω_I::Vector{T}
    ω_F::Vector{T}
    ag_share::Vector{T} # want to make sure it's the same as in the data
    gdppc::Vector{T} # this is not in logs, which shouldn't matter
    manuf_share_GDP::Vector{T} # need not be the same
    pop_share::Vector{T} # should recover data perfectly
    T::Vector{T}
    Q::Vector{T}
    HrM::Vector{T}
    HrA::Vector{T}
    ag_wages::Vector{T}
    manuf_wages::Vector{T}
    PV2::param
end


# stores intermediate output
mutable struct CO4{T}
    bwguess_F::Vector{T}
    bwguess_I::Vector{T}
    pi_ia::Vector{T}
    pi_fa::Vector{T}
    pi_im::Vector{T}
    pi_fm::Vector{T}
    ω_I::Vector{T}
    ω_F::Vector{T}
    ag_wages::Vector{T}
    manuf_wages::Vector{T}
    region_share_I::Vector{T}
    region_share_F::Vector{T}
    region_Num_I::Vector{T}
    region_Num_F::Vector{T}
    V::Vector{T}
    N::Vector{T}
end

# need to feed in data using this structure
mutable struct Data_Input_Calibration{T}
    pop_share::Vector{T}
    tot_pop::T
    ag_share::Vector{T}
    gdppc::Vector{T}
end



# Inner function to solve for manufacturing wages

function Find_Manuf_Wages(gdppc::Array{Float64,1}, pm::param,
        ω_I::Array{Float64,1}, ω_F::Array{Float64,1}, wguess_A::Array{Float64,1},
        pi_i::Array{Float64,1}, pi_f::Array{Float64,1})



    @unpack θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ, fE=pm

     #  parameter combinations
    Γ_θ=gamma(1-1/θ)
    Denom=(1-(1-δ)*(1-β)/(ρ))
    L=length(wguess_A)
    wguess_M=ones(L)
    residvec=ones(L)

      for r in 1:L
            resid2(x)=abs( gdppc[r]-
            Γ_θ *( ((1/Denom)*(ω_I[r]*(ϕ_ia*wguess_A[r]^θ+ϕ_im*(x)^θ)^(1/θ)*(1-pi_i[r])

                      + ω_F[r]*(ϕ_fa*wguess_A[r]^θ+ϕ_fm*(x)^θ)^(1/θ)*(1-pi_f[r]))

                      +(1/γ)*(ω_I[r]*(ϕ_ia*wguess_A[r]^θ+ϕ_im*(x)^θ)^(1/θ)*(pi_i[r])

                      + ω_F[r]*(ϕ_fa*wguess_A[r]^θ+ϕ_fm*(x)^θ)^(1/θ)*(pi_f[r])))        ) )

            j=optimize(resid2, 0, 100000)
            wguess_M[r]=Optim.minimizer(j)
            residvec[r]=Optim.minimum(j)

    end
    max_problem=findmax(abs.(residvec)./wguess_M)[1]
    if max_problem>0.5
       # println("manuf error too big")
    end

    return wguess_M, residvec
end

# Recovers average wages as part of steady state
function recover_wbar(Data_test::Data_Input_Calibration, pm::param, wbarguess_F::Array{Float64,1}, wbarguess_I::Array{Float64,1}, print=0)
    @unpack pop_share, tot_pop, ag_share, gdppc= Data_test
    @unpack θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ, fE=pm
    Γ_θ=gamma(1-1/θ)
    𝜈=(1-β)/(ρ-1)
    X=(𝜈+λ*𝜈/(1-λ))
    if β==0.0
        Denom=(1-(1-δ)*(1-β)/(ρ))
    else
        Denom=(1/β)*(1-(1-δ)*(1-β)/(ρ)) 
    end
    L=length(pop_share) # store length for loops

    iter1=1 # for Wbar loop
    tol1=10

    V=pop_share # feed in pop shares as an initial guess
    Normalizer=sum(V .^ϵ)^(1/ϵ)
    V=V./Normalizer
    TV=ones(L)

    bwguess_F=wbarguess_F # preallocate wage vector
    bwguess_I=wbarguess_I # preallocate wage vector

    T_bwguess_F=ones(L)
    T_bwguess_I=ones(L)

    pi_ia=zeros(L) 
    pi_fa=zeros(L)
    pi_im=zeros(L) 
    pi_fm=zeros(L)

    ω_I=zeros(L)
    ω_F=zeros(L)
    N=zeros(L)

    region_share_I=zeros(L)
    region_share_F=zeros(L)

    ag_wages=zeros(L)
    denom_A=zeros(L)
    manuf_wages=ones(L)
    residvec=ones(L)

    maxtol=0.1


     while tol1>maxtol && iter1<2000 # looping over bar w

        # Get location values V
        iterV=1 # for V loop
        tolV=10
         while tolV>0.0001 && iterV<1000
           denom_I1=sum((V.*bwguess_I).^ϵ)
           denom_F1=sum((V.*bwguess_F).^ϵ)

            TV .= (pop_share./ (χ.*bwguess_I.^ϵ./denom_I1.+(1-χ).*bwguess_F.^ϵ./denom_F1 )).^(1/ϵ)

            tolV=100*findmax(abs.(TV-V)./V)[1]
            V=(0.1.*TV+0.9.*V)
            V=V./(sum(V .^ϵ)^(1/ϵ))
            TV=ones(L)
            iterV=iterV+1
            if iterV>999
            println("Max iters exceeded in V")
            end

        end

        # Get regional employment shares by type 
        region_share_I .= (V.*bwguess_I).^ϵ./sum((V.*bwguess_I).^ϵ)
        region_share_F .= (V.*bwguess_F).^ϵ./sum((V.*bwguess_F).^ϵ)

         ω_I .=region_share_I .*χ     .*tot_pop ./ (region_share_I .* χ .* tot_pop .+ region_share_F .* (1-χ) .*tot_pop)
         ω_F .=region_share_F .*(1-χ) .*tot_pop ./ (region_share_I .* χ .* tot_pop .+ region_share_F .* (1-χ) .*tot_pop)

        # Calculate ag wages
        denom_A .=ω_I .*ϕ_ia .* (1 ./bwguess_I).^θ .+ ω_F .*ϕ_fa .*(1 ./bwguess_F).^θ
        ag_wages .=(ag_share./denom_A ).^(1/θ)

        # Calculate manufacturing wages
        pi_ia .=ϕ_ia .*(ag_wages ./bwguess_I).^θ
        pi_fa .=ϕ_fa .*(ag_wages ./bwguess_F).^θ
        mw=Find_Manuf_Wages(gdppc, pm,ω_I, ω_F, ag_wages, pi_ia, pi_fa)
        manuf_wages=mw[1]
        residvec=mw[2]

        # Get sectoral employment share of industrial workers
        pi_im .=ϕ_im .*(manuf_wages ./bwguess_I).^θ
        pi_fm .=ϕ_fm .*(manuf_wages ./bwguess_F).^θ


        # Update guess for w_bar
        T_bwguess_F .= (ϕ_fm .*manuf_wages.^θ .+ ϕ_fa .*ag_wages.^θ).^(1/θ)
        T_bwguess_I .= (ϕ_im .*manuf_wages.^θ .+ ϕ_ia .*ag_wages.^θ).^(1/θ)

        # Tolerances
        tol1=max( 100*findmax(abs.(T_bwguess_F-bwguess_F)./bwguess_F)[1],
        100*findmax(abs.(T_bwguess_I-bwguess_I)./bwguess_I)[1] )

       # Updatimg
        bwguess_F=(0.1.*T_bwguess_F+0.9 .*bwguess_F)#./bwguess_F[1]#./Q_c
        bwguess_I=(0.1.*T_bwguess_I+0.9 .*bwguess_I)#./bwguess_F[1]#./Q_c

        iter1=iter1+1
        if iter1>999
            println("Max iters exceeded main loop")
        end


    end #this ends the main for-loop and recovers Wbar.
    
    region_Num_I=zeros(L)
    region_Num_F=zeros(L)
    region_Num_I.=region_share_I.*χ.*tot_pop
    region_Num_F.=region_share_F.*(1-χ).*tot_pop
    HrM=zeros(L)
    HrM .= Γ_θ .* region_Num_I .* ϕ_im .* (manuf_wages ./bwguess_I) .^(θ-1)+
               Γ_θ .* region_Num_F .* ϕ_fm .* (manuf_wages ./bwguess_F) .^(θ-1)

    NoEntry=zeros(L)
    N=zeros(L)
    if β==0.0
        N.=((1-β)/(ρ)*(1-δ)^λ/fE.*HrM./Denom).^(1/(1-λ))
    else
        N.=((1-β)/(ρ*β)*(1-δ)^λ/fE.*HrM./Denom).^(1/(1-λ))
    end

    if print==1
       println(iter1," iterations ", tol1, " Delta")
    println("Largest residual for manufacturing wages is ",findmax(abs.(residvec))[1])
    end

    C1=CO4(bwguess_F,bwguess_I,pi_ia,pi_fa,pi_im,pi_fm,ω_I,ω_F ,ag_wages,manuf_wages,region_share_I, region_share_F,
        region_Num_I, region_Num_F, V,N)
    return  C1
end

# Recovers rest of  fundamentals
function recover_fundamentals(Data_test::Data_Input_Calibration, pm::param, C1::CO4,print=0)

    @unpack pop_share, tot_pop, ag_share, gdppc= Data_test
    @unpack θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ,η,ρ_a, fE=pm
    @unpack bwguess_F, bwguess_I, pi_ia, pi_fa, pi_im, pi_fm, ω_I, ω_F, ag_wages, manuf_wages,region_share_I, region_share_F, V=C1

    L=length(pop_share)
    Γ_θ=gamma(1-1/θ)
    𝜈=(1-β)/(ρ-1)
    X=(𝜈+λ*𝜈/(1-λ))  
    if β==0.0
        Denom1=(1-(1-δ)*(1-β)/(ρ))  
    else
        Denom1=(1/β)*(1-(1-δ)*(1-β)/(ρ))  
    end
    Denom_manuf=(1-(1-δ)*𝜈*(1-β)/(𝜈+1-β))  

    ζ=((ρ-1)/ρ*(1-β)/β)^(1-β)*((1-β)/(ρ*β)*(1-δ)^λ/fE)^𝜈

    # recover other fundamentals: preallocation
    HrM=zeros(L)
    HrA=zeros(L)
    z=zeros(L)
    q=zeros(L)
    Q_c=1
    T=zeros(L)
    Q=zeros(L)

    # Get Omega terms to recover Q
    Ω_M=1
    Ω_A=1
    Q_c=1
    Z=1

    # Sectoral human capital
      HrA .=Γ_θ .* ω_I.*pop_share.*tot_pop.*ϕ_ia.*(ag_wages./bwguess_I).^(θ-1)+
        Γ_θ.*ω_F.*pop_share.*tot_pop.*ϕ_fa.*(ag_wages./bwguess_F).^(θ-1)

       HrM .= Γ_θ.*ω_I.*pop_share.*tot_pop.*ϕ_im.*(manuf_wages./bwguess_I).^(θ-1)+
        Γ_θ.*ω_F.*pop_share.*tot_pop.*ϕ_fm.*(manuf_wages./bwguess_F).^(θ-1)


    #
    denom_z=sum(ag_wages.^σ.*HrA.^(1-(1-σ)*(1-γ)) )
    z.=(ag_wages.^σ .*HrA.^(1-(1-σ)*(1-γ))./ denom_z).^(1/(σ-1))
    
    # Solve for alpha to be consistent with BGP
    α_r=(1/γ)*(sum(ag_wages.*HrA))/sum(gdppc.*pop_share.*tot_pop)


    PV2=param(θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α_r, ɷ, ϱ, ρ, γ, β, λ, δ,η,ρ_a, fE)

    # Get relative technologies
    denom_Q=sum(manuf_wages.^σ.*HrM.^(1-(σ-1)*𝜈/(1-λ)) )

    q .=(manuf_wages.^σ.*HrM.^(1-(σ-1)*𝜈/(1-λ))./ denom_Q).^(1/(σ-1))

    return Fundamentals=(V=V, Param=PV2)
end



# create the steady state struct
mutable struct SteadyState{T}
    # Data and Parameters
    data::Data_Input_Calibration{T}
    pm::param
    μe::Matrix{T}
    τ::Matrix{T}
    # Equilibrium Objects
    N::Vector{T}
    Q::Vector{T}
    T::Vector{T}
    V::Vector{T}
    ω_I::Vector{T}
    ω_F::Vector{T}
    wb_F::Vector{T}
    wb_I::Vector{T}
    ag_wages::Vector{T}
    manuf_wages::Vector{T}
    Ψ::Vector{T}
    L_I::Vector{T}
    L_F::Vector{T}
    pi_ia::Vector{T}
    pi_fa::Vector{T}
    HrA::Vector{T}
    HrM::Vector{T}
    HrP::Vector{T}
    PrrM::Vector{T}
    PrrA::Vector{T}
    Prm::Vector{T}
    Pra::Vector{T}
    # Solution Algorithm Params
    maxtol::Float64 # stores biggest percent deviation 
    # Scratch Vectors
    _scratch1::Vector{T}
    _scratch2::Vector{T}
    _scratch3::Vector{T}
    _scratch4::Vector{T}
end

# instantiate the steady state
function SteadyState(data::Data_Input_Calibration, pm::param,
                TradeCosts, MobilityCosts)
    @unpack ϵ = pm

    L           = length(data.pop_share)
    μe          = MobilityCosts .^ ϵ
    τ           = TradeCosts
    N           = ones(L)
    Q           = ones(L)
    T           = ones(L)
    V           = ones(L)
    ω_I         = ones(L)
    ω_F         = ones(L)
    wb_F        = ones(L)
    wb_I        = ones(L)
    ag_wages    = ones(L)
    manuf_wages = ones(L)
    Ψ           = ones(L)
    L_I         = ones(L)
    L_F         = ones(L)
    pi_ia       = ones(L)
    pi_fa       = ones(L)
    HrA         = ones(L)
    HrM         = ones(L)
    HrP         = zeros(L)
    PrrM        = ones(L)
    PrrA        = ones(L)
    Prm         = ones(L)
    Pra         = ones(L)
    maxtol      = 0.05
    _scratch1   = ones(L)
    _scratch2   = ones(L)
    _scratch3   = ones(L)
    _scratch4   = ones(L)

    return SteadyState(data, pm, μe, TradeCosts, N, Q, T, V, ω_I, ω_F, wb_F,
                       wb_I, ag_wages, manuf_wages, Ψ, L_I, L_F, pi_ia, pi_fa,
                       HrA, HrM, HrP, PrrM, PrrA, Prm , Pra , maxtol, _scratch1,
                       _scratch2, _scratch3, _scratch4)
end

#=
Here are some helper functions
=#

innertol(s::SteadyState)    = 0.01
len(s::SteadyState)         = length(s.V)

function err(TV, V)
    mapreduce(max, TV, V) do new, old
        abs(new - old)/abs(old)
    end
end

function solve_V!(s::SteadyState)
    @unpack μe, L_I, wb_I, Ψ, V,maxtol  = s
    @unpack ϵ   = s.pm

    # Fixed Point Accumulators
    iterV   = 1
    tolV    = 10.0

    # Rename scratch buffers
    wdf = s._scratch1
    den = s._scratch2
    tmp = s._scratch3
    TV  = s._scratch4

    tmp .= V.^ϵ
    V   .= V./(sum(tmp)^(1/ϵ))

    # Calculate wdf at the top of the loop
    wdf .= (wb_I ./ Ψ).^ϵ

    while tolV > innertol(s) && iterV<1000
        # Calculate denominator
        tmp .= wdf .* V.^ϵ
        mul!(den, μe, tmp)

        @inbounds for r in 1:len(s)
            # Now compute the sum in the denominator
            d = 0.0
            c = wdf[r]
            @inbounds @simd for k in 1:len(s)
                d += L_I[k] * μe[k,r] * c / den[k]
            end

            # Compute the next value step in the contraction
            TV[r] = (L_I[r]/d)^(1/ϵ)
        end

        # Compute relative error
        tolV =  err(TV, V)

        # Update V (inplace)
        V   .= 0.4 .* TV .+ 0.6 .*V
        tmp .= V.^ϵ
        V   .= V./(sum(tmp)^(1/ϵ))

        # Update iteration counter
        iterV += 1

        if iterV > 999
            println("Max iters exceeded in V")
            #maxtol=tolV
        end
    end
    #@show iterV tolV

    return
end

function solve_ω!(s::SteadyState)
    @unpack L_I, L_F, ω_I, ω_F, wb_F, Ψ, V, μe, maxtol   = s
    @unpack pop_share                       = s.data
    @unpack ϵ = s.pm

    # Rename scratch vectors
    Farmvec = s._scratch1
    T_ω_I   = s._scratch2
    temp    = s._scratch3
    acc=s._scratch4
    tol_omega=10.0
    iter_omega = 1
    while tol_omega > 0.1 && iter_omega<1000 # third loop for omega
        # Recover V, get new L_F, update omega
        L_I .= ω_I.*pop_share 
        # Update Values in place
        solve_V!(s)

        # Calculate vector of farm population
        fill!(Farmvec, 0.0)
        @inbounds for j in 1:len(s)
            c = (wb_F[j] / Ψ[j] * V[j])^ϵ
            @inbounds @simd for k in 1:len(s)
                Farmvec[k] += μe[k,j] * c
            end
        end

        # New farmer shares
        acc .= 1 ./Farmvec.* (1 .- ω_I) .* pop_share

        @inbounds for r in 1:len(s)
            val = 0.0
            c   = (V[r] * wb_F[r] / Ψ[r])^ϵ
            @inbounds @simd for k in 1:len(s)
                acc2  = μe[k, r] * c
                acc2 *= acc[k]
                val  += acc2
            end
            L_F[r] = val
        end

        # Update Omega
        @inbounds for r in 1:len(s)
            if 1-L_F[r]/pop_share[r] > 0
                T_ω_I[r]=1-L_F[r]/pop_share[r]
            else
                T_ω_I[r]=0.0
            end
        end

        # Compute error in omega
        tol_omega = 100 * err(T_ω_I, ω_I)

        # Update industrialist share
        ω_I .= (0.1 .* T_ω_I .+ 0.9 .* ω_I)

        iter_omega=iter_omega+1
        if iter_omega>999
                 if s.maxtol==0.1 
            println("Max iters exceeded omega loop, ", tol_omega)
                end
        s.maxtol=deepcopy(tol_omega)
        end
    end

    # Update farm shares
    ω_F.=L_F./pop_share

    return
end


function Find_Manuf_Wages(s::SteadyState, args...)
    @unpack gdppc    = s.data
    @unpack ω_I, ω_F, pm = s
    return Find_Manuf_Wages(gdppc, pm, ω_I, ω_F, args...)
end


function solve_w!(s::SteadyState)

    @unpack ϕ_ia, ϕ_fa,ϕ_fm,ϕ_im, θ                     = s.pm
    @unpack ag_share                                    = s.data
    @unpack ω_I, ω_F, wb_I, wb_F, L_F, ag_wages, manuf_wages = s
    @unpack pi_ia, pi_fa                                = s

    # Rename scratch buffers
    T_wb_F = s._scratch1
    T_wb_I = s._scratch2
    tol_w=10.0
    iter_w=1
    while tol_w>0.1 && iter_w<1000 # second loop for w

        # restart counter each time
        tol_omega=10.0
        iter_omega=1

        #Placeholder
        solve_ω!(s)

        # Calculate ag wages
        ag_wages .= ag_share
        ag_wages.*= 1 ./(ω_I .* ϕ_ia .* (1 ./wb_I).^θ
                     .+ ω_F .* ϕ_fa .* (1 ./wb_F).^θ )
        ag_wages .= ag_wages .^ (1/θ)

        # Calculate manufacturing wages, can recycle old fcn
        pi_ia .= ϕ_ia .*(ag_wages ./wb_I).^θ
        pi_fa .= ϕ_fa .*(ag_wages ./wb_F).^θ
        mw     = Find_Manuf_Wages(s, ag_wages, pi_ia, pi_fa)
        manuf_wages .= mw[1]

        # finally we update the bar w guesses
        T_wb_F .= (ϕ_fm .*manuf_wages.^θ .+ ϕ_fa .*ag_wages.^θ).^(1/θ)
        T_wb_I .= (ϕ_im .*manuf_wages.^θ .+ ϕ_ia .*ag_wages.^θ).^(1/θ)

        # Tolerances
        e1    = 100 * err(T_wb_F, wb_F)
        e2    = 100 * err(T_wb_I, wb_I)
        tol_w = max( e1, e2 )

        # Updating
        wb_F .=(0.1.*T_wb_F+0.9 .*wb_F)
        wb_I .=(0.1.*T_wb_I+0.9 .*wb_I)
        iter_w += 1
        #println(iter_w)
        if iter_w>999
            println("Max iters exceeded w loop")
        end
    end

end


function solve_P!(s::SteadyState)
    @unpack θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ, fE=s.pm
    @unpack PrrA, PrrM, ag_wages, manuf_wages, τ, HrA, HrP = s
    tol=10.0
    iter=1

    # Rename Scratch Buffers
    dena     = s._scratch1
    denm     = s._scratch4

    T_PrrA  = s._scratch2
    T_PrrM  = s._scratch3


    while tol > 0.1 && iter < 1000
        # Calculate Denominator vectors
        @inbounds @simd for j in 1:len(s)
            d1 = 0.0
            d2 = 0.0
            for m in 1:len(s)
                c   = τ[m,j]^(1-σ)
                d1 += c * PrrA[m]^(1-σ)
                d2 += c * PrrM[m]^(1-σ)
            end
            dena[j] = d1
            denm[j] = d2
        end

        # Calculate the big sum
        @inbounds @simd for r in 1:len(s)
            v1 = 0.0
            v2 = 0.0
            @inbounds @simd for j in 1:len(s)
                if β==0.0
                    c     = (HrA[j] * ag_wages[j]/γ + manuf_wages[j]*HrP[j])
                else
                    c     = (HrA[j] * ag_wages[j]/γ + manuf_wages[j]*HrP[j]/β)
                end
                acc1  = τ[r,j]^(1-σ)/dena[j] * c
                v1   += acc1

                acc2  = τ[r,j]^(1-σ)/denm[j] * c
                v2   += acc2
            end
            T_PrrA[r]  = ( γ/(ag_wages[r]*HrA[r]) * α * v1 )^(1/(σ-1))
            if β==0.0
                T_PrrM[r]  = ( 1/(manuf_wages[r]*HrP[r]) * (1-α) * v2 )^(1/(σ-1))
            else
                T_PrrM[r]  = ( β/(manuf_wages[r]*HrP[r]) * (1-α) * v2 )^(1/(σ-1))
            end
        end

           @inbounds @simd for r in 1:len(s)
            end

        # Update own region prices
        tol     = 100 * max(err(T_PrrA, PrrA), err(T_PrrM, PrrM))
        PrrA   .= (0.75 .* T_PrrA + 0.25 .* PrrA)./PrrA[1]
        PrrM   .= (0.75 .* T_PrrM + 0.25 .* PrrM)./PrrM[1]

        # Update Iteration Count
        iter += 1
        if iter>999
            println("Max iters exceeded M loop")
        end
    end
end


function solve_Ψ!(s::SteadyState)

    @unpack pop_share, tot_pop, gdppc = s.data
    @unpack ag_wages, manuf_wages, wb_I, wb_F, ω_I, ω_F = s
    @unpack HrA, HrM, HrP, PrrM, PrrA, Prm, Pra, τ, Ψ   = s
    @unpack θ, γ, ρ, δ, β,  σ, α                        = s.pm
    @unpack ϕ_ia, ϕ_im, ϕ_fa, ϕ_fm                      = s.pm

    Γ_θ = gamma(1-1/θ)
    𝜈   = (1-β)/(ρ-1)
    Ψ.=Ψ./Ψ[1]
    # Rename scratch buffer
    T_Ψ= s._scratch1
    tol_Ψ=10.0
    iter_Ψ=1
    while tol_Ψ > 0.1 && iter_Ψ < 1000 # outer loop for Ψ

        # Solve fixed point for w
        solve_w!(s)

        # recover HrA, HrM
        HrA .= Γ_θ .* ω_I.*pop_share.*tot_pop.*ϕ_ia.*(ag_wages./wb_I).^(θ.-1).+
               Γ_θ.*ω_F.*pop_share.*tot_pop.*ϕ_fa.*(ag_wages./wb_F).^(θ.-1)

        HrM .= Γ_θ.*ω_I.*pop_share.*tot_pop.*ϕ_im.*(manuf_wages./wb_I).^(θ.-1).+
               Γ_θ.*ω_F.*pop_share.*tot_pop.*ϕ_fm.*(manuf_wages./wb_F).^(θ.-1)

        if β==0.0
            HrP .= HrM.*(ρ)./((ρ-1+δ)*(1-β))
        else
            HrP .= HrM.*(ρ*β)./(ρ*β+(ρ-1+δ)*(1-β))
        end
        
        # Update αgricultural employment share parameter
        α_r=(1/γ)*(sum(ag_wages.*HrA))/sum(gdppc.*pop_share.*tot_pop)
        s.pm = param(s.pm, α = α_r)

        # Get PrrM
        solve_P!(s)

        # Tolerance
        @inbounds for r in 1:len(s)
            Prm[r] = 0.0
            Pra[r] = 0.0
            @inbounds @simd for k in 1:len(s)
                Prm[r] += τ[k,r]^(1-σ) * PrrM[k]^(1-σ)
                Pra[r] += τ[k,r]^(1-σ) * PrrA[k]^(1-σ)
            end

        end

            Prm .= Prm .^(1/(1-σ))
            Pra .= Pra .^(1/(1-σ))
        # Aggregate income
        T_Ψ     .= (Pra./α_r).^α_r.*(Prm./(1-α_r)).^(1-α_r)
        T_Ψ     .= T_Ψ./T_Ψ[1]
        tol_Ψ    = 100 * err(T_Ψ, Ψ)

        # Updating
        Ψ       .= (0.5.*T_Ψ+0.5 .*Ψ)
        Ψ       .= Ψ./Ψ[1]
        iter_Ψ  += 1

        if iter_Ψ>999
            println("Max iters exceeded main loop")
        end
    end
end



function model_setup(data, pm, wbarguess_F, wbarguess_I,ωguess_I,Ψguess,
                Vguess, TradeCosts, MobilityCosts)

    @unpack θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ, fE=pm

    s = SteadyState(data, pm, TradeCosts, MobilityCosts)

    s.wb_I   = copy(wbarguess_I)
    s.wb_F   = copy(wbarguess_F)
    s.ω_I    = copy(ωguess_I)
    s.Ψ      = copy(Ψguess)
    s.V      = copy(Vguess)

    return s
end



function solve!(s::SteadyState)

    @unpack θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ,η,ρ_a, fE = s.pm
    @unpack manuf_wages, ag_wages, HrP, HrA, HrM, PrrA, PrrM, Prm, Pra, maxtol = s
    @unpack N, Q, T, L_I, L_F, ω_I  = s
    @unpack pop_share, tot_pop = s.data

    solve_Ψ!(s)

    if β==0.0
        N   .=  ((1-β)/(ρ*fE).*HrP).^(1/(1-λ))
    else
        N   .=  ((1-β)/(ρ*β*fE).*HrP).^(1/(1-λ))
    end

    𝜈=(1-β)/(ρ-1)

        Q.= manuf_wages./(PrrM.*HrP.^𝜈.*N.^(λ*𝜈))


    T   .= (ag_wages./(γ.*Q.*PrrA)).^(1/(1-γ)).*HrA
    L_I .= ω_I.*pop_share.*tot_pop
    L_F .= (1 .-ω_I).*pop_share.*tot_pop
end



function Recover_Fundamentals_NEW(args...; kwargs...)

    s = model_setup(args...;kwargs...)
    solve!(s)
    return s
end



#=
Next Section: Simulation Codes
=#


struct wage_clearing_output9
    ag_wage::Array{Float64,1}
    manuf_wage::Array{Float64,1}
    LNI_p::Array{Float64,1} # native industrial today
    LNF_p::Array{Float64,1} # native rural today
    LRI_p::Array{Float64,1} # refugee industrial today
    LRF_p::Array{Float64,1} # refugee rural today
    N_firms::Array{Float64,1} #
    N_indicator::Array{Float64,1} # refugee rural today
    mi::Array{Float64,2} # these are matrices now--how many moved in from column region to row region
    mf::Array{Float64,2} # these are matrices now
    Psivec::Array{Float64,1}
    HrP::Array{Float64,1}
    HrA::Array{Float64,1}

end


function Get_Mobility(mni::Array{Float64,2},mri::Array{Float64,2},
        mnf::Array{Float64,2},mrf::Array{Float64,2},
        ψ::Float64,μe::Array{Float64,2},
        VWiΨ::Array{Float64,1},VWfΨ::Array{Float64,1},
        μVWiΨ::Array{Float64,1},μVWfΨ::Array{Float64,1},
        LNI::Array{Float64,1},LRI::Array{Float64,1},
        LNF::Array{Float64,1},LRF::Array{Float64,1},
        mni_v::Array{Float64,1},mri_v::Array{Float64,1},
        mnf_v::Array{Float64,1},mrf_v::Array{Float64,1},
        L::Int, c::Float64)
                fill!(mni_v, 0.0)
                fill!(mri_v, 0.0)
                fill!(mnf_v, 0.0)
                fill!(mrf_v, 0.0)
    @inbounds  @simd for r in 1:L
                iv=VWiΨ[r]
                fv=VWfΨ[r]
                @inbounds   @simd for k in 1:L
            c =μe[k,r]
                    mni[k,r]=LNI[k]*ψ*c*iv/μVWiΨ[k]::Float64
                    mri[k,r]=LRI[k]*ψ*c*iv/μVWiΨ[k]::Float64
                    mnf[k,r]=LNF[k]*ψ*c*fv/μVWfΨ[k]::Float64
                    mrf[k,r]=LRF[k]*ψ*c*fv/μVWfΨ[k]::Float64
                     mni_v[r]+=mni[k,r]
                     mri_v[r]+=mri[k,r]
                     mnf_v[r]+=mnf[k,r]
                     mrf_v[r]+=mrf[k,r]
                end
            end


end


function Get_T(L::Int64,c::Float64,
                Prms::Float64, Pras::Float64,
                tm::Array{Float64,1}, ta::Array{Float64,1},
                PrrmSig::Array{Float64,1},PrraSig::Array{Float64,1},
                PrmSig::Array{Float64,1}, PraSig::Array{Float64,1},
                PY::Array{Float64,1},τs::Array{Float64,2},σ::Float64)
        fill!(tm, 0.0)
        fill!(ta, 0.0)

            @inbounds @simd for j in 1:L
                c=PY[j]::Float64
                Prms=PrmSig[j]::Float64 #exp((1-σ)*log(Prm[j]))
                Pras=PraSig[j]::Float64 #exp((1-σ)*log(Prm[j]))
            @inbounds @simd for r in 1:L
                tm[r]+=τs[r,j]*PrrmSig[r]*c/(Prms)::Float64
                ta[r]+=τs[r,j]*PrraSig[r]*c/(Pras)::Float64
                end

            end
end


function Wage_Clearing_endogV(wguess_A1::Array{Float64,1}, wguess_M1::Array{Float64,1},
        T::Array{Float64,1}, V::Array{Float64,1}, Q::Array{Float64,1},N::Array{Float64,1},
        LNI::Array{Float64,1},LNF::Array{Float64,1}, LRI::Array{Float64,1}, LRF::Array{Float64,1},
        Immig_Shock_I::Array{Float64,1}, Immig_Shock_F::Array{Float64,1}, Pv::param,
        β_H::Float64,η::Float64,
        TradeCosts::Array{Float64,2}, MobilityCosts::Array{Float64,2},Ψguess1::Array{Float64,1};
        iters=1000, tol=0.1, convergence_speed=0.9, convergence_psi=0.5)
     @unpack  θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ, fE=Pv
    wguess_A=deepcopy(wguess_A1)
    wguess_M=deepcopy(wguess_M1)
    L=length(wguess_A)

    # a few messy terms
    Γ_θ=gamma(1-1/θ)
    𝜈=(1-β)/(ρ-1) #short run scale elasticity
    if β==0.0
        pm_const = ((ρ-1)/ρ*(1-β))^(β-1)
    else
        pm_const = ((ρ-1)/ρ*(1-β)/β)^(β-1) #* (1/β)^(β)
    end

    diff_Ψ=10.0 # ratio of input guess to output guess
    iter_Ψ=1

    diff_w=10.0 # ratio of input guess to output guess
    iter_w=1

    ################################################
    ######## Preallocate vectors
    ################################################

        # Aggregate population
    LNI_agg=sum(LNI) # native industrialists
    LNF_agg=sum(LNF) # native farmers
    LRI_agg=sum(LRI) # refugee industrialists
    LRF_agg=sum(LRF) # refugee farmers

    # Sectoral labor supply
    HrA=zeros(L)
    HrM=zeros(L)

    # Manufacturing labor market clearing
    HrP=zeros(L)
    Pos_Inv=zeros(L)

    # Regional prices
    Prrm=zeros(L)
    Prra=zeros(L)
    Prm=zeros(L)
    Pra=zeros(L)
    # will store prices^(1-\sigma)
    PrrmSig=zeros(L)
    PrraSig=zeros(L)
    PrmSig=zeros(L)
    PraSig=zeros(L)
    PY=zeros(L)

    # Vectors for wage update
    wprime_M=zeros(L)
    wprime_A=zeros(L)

    # N update
    Nprime=zeros(L)

    # w_bar terms
    wbar_I=zeros(L)
    wbar_F=zeros(L)
    w_norm_A=zeros(L)
    w_norm_M=zeros(L)

    # moving terms
    mi=zeros(L,L) # these are matrices now--how many moved in from column region to row region
    mf=zeros(L,L) # these are matrices now

    mni=zeros(L,L)
    mri=zeros(L,L)
    mnf=zeros(L,L)
    mrf=zeros(L,L)

    mni_v=zeros(L)
    mri_v=zeros(L)
    mnf_v=zeros(L)
    mrf_v=zeros(L)


    # Instead, make initial guess previous labor allocation
    LNI_p=copy(LNI)
    LNF_p=copy(LNF)
    LRI_p=copy(LRI)
    LRF_p=copy(LRF)

    # Industrialists/Farmers are sum of natives, refugees
    LI_p=LNI_p+LRI_p
    LF_p=LNF_p+LRF_p

    NoEntry=zeros(L)
    Nactive=zeros(L)

    μ=MobilityCosts
    μe=μ.^ϵ # I only want to do this once
    τ=TradeCosts
    τs=τ.^(1-σ)
    τst=τs'
    # preallocate matrices and vectors for wage clearing
    tpm=zeros(L,L)
    tpa=zeros(L,L)
    tm=zeros(L)
    ta=zeros(L)
            c=1.0
            Prms=1.0
            Pras=1.0



    Ψguess=deepcopy(Ψguess1)
    Ψguess=Ψguess./Ψguess[1]
    T_Ψguess=ones(L)

    V_p=copy(V)
    VWiΨ=zeros(L)
    VWfΨ=zeros(L)
    μVWiΨ=zeros(L)
    μVWfΨ=zeros(L)

    MV=ones(L)

    ################################################
    ######## Main Loop
    ################################################
   
    while  diff_Ψ>tol && iter_Ψ<iters # this loops over Ψ till it finds a fixed point
        diff_w=10 # ratio of input guess to output guess
        iter_w=1

          while  diff_w>tol && iter_w<iters # this loops over Ψ till it finds a fixed point

       wbar_I .= (ϕ_ia .* wguess_A.^θ .+ ϕ_im .* wguess_M.^θ).^(1/θ)
       wbar_F .= (ϕ_fa .* wguess_A.^θ .+ ϕ_fm .* wguess_M.^θ).^(1/θ)

        # Adjust amenities for population
        V_p=V.*((LI_p.+LF_p).^(-β_H)) 
            
        VWiΨ.=(wbar_I./Ψguess.*V_p).^ϵ
      
         mul!(μVWiΨ, μe, VWiΨ)

        VWfΨ.=(wbar_F./Ψguess.*V_p).^ϵ
    
        mul!(μVWfΨ, μe, VWfΨ)

        # Calculate mobility
         Get_Mobility(mni,mri,mnf,mrf,ψ,μe,VWiΨ,VWfΨ,μVWiΨ,μVWfΨ,LNI,LRI,LNF,LRF,
                mni_v,mri_v,mnf_v,mrf_v,
                L,c)

                # Calculate spatial labor supply
                # Industrialists
                LNI_p .= (1-ψ) .* LNI .+ mni_v 
                LRI_p .= (1-ψ) .* LRI .+ mri_v .+ Immig_Shock_I
                LI_p .= LNI_p .+ (η * LRI_p)

                # Farmers
               LNF_p .= (1-ψ) .* LNF .+ mnf_v
                LRF_p .= (1-ψ) .* LRF .+ mrf_v .+ Immig_Shock_F
               LF_p .= LNF_p .+ (η * LRF_p)

        # Calculate sectoral labor supply
         HrA .= Γ_θ .* LI_p .* ϕ_ia .* (wguess_A./wbar_I) .^ (θ-1) .+
                Γ_θ .* LF_p .* ϕ_fa .* (wguess_A./wbar_F) .^(θ-1)

        HrM .= Γ_θ .* LI_p .* ϕ_im .* (wguess_M ./wbar_I) .^(θ-1).+
               Γ_θ .* LF_p .* ϕ_fm .* (wguess_M ./wbar_F) .^(θ-1)


        # Define indicator where the free entry condition is slack
       @inbounds @simd for r in 1:L
            if HrM[r] < (ρ - (1-β)) / (1-β) * fE * (1-δ)*(N[r])^(1-λ)
                NoEntry[r] = 1 # slack
            else
                NoEntry[r]= 0 # binding
            end
        end

        # Calculate number of production workers
        if β==0.0
            HrP .= NoEntry.* (ρ/ (ρ - (1-β))) .* HrM .+ (1 .-NoEntry).* (HrM .+ fE .*(1-δ).* (N).^(1-λ) )
        else
            HrP .= NoEntry.* (ρ*β / (ρ - (1-β))) .* HrM .+ (1 .-NoEntry).*  β .* (HrM .+ fE .*(1-δ).* (N).^(1-λ) )
        end


        # Calculate number of active plants
        if β==0.0
            Nactive .=  NoEntry.* (1-δ).*N + (1 .-NoEntry).* ((1-β)/(ρ)) .* 1/fE .* (N).^(λ) .* HrP
        else
            Nactive .=  NoEntry.* (1-δ).*N + (1 .-NoEntry).* ((1-β)/(ρ*β)) .* 1/fE .* (N).^(λ) .* HrP
        end

        # Get regional prices
        if β==0.0
            Prrm .= wguess_M .* Nactive.^(-1*(1-β)/(ρ-1)) .* Q.^(-1) .* pm_const
        else
            Prrm .= wguess_M .* Nactive.^(-1*(1-β)/(ρ-1)) .* Q.^(-1) .* pm_const./β
        end
        Prra .= wguess_A .* (HrA./T).^(1-γ) .* Q.^(-1)
        PrrmSig.=Prrm.^(1-σ)
        PrraSig.=Prra.^(1-σ)

        Prm.= (τst*(PrrmSig)).^(1/(1-σ))
        Pra.= (τst*(PrraSig)).^(1/(1-σ))


        # Aggregate income

        T_Ψguess.=(Pra./α).^α.*(Prm./(1-α)).^(1-α)
        w_norm_A.=wguess_A
        w_norm_M.=wguess_M
        if β==0.0
            PY.=w_norm_A .* HrA./γ .+ w_norm_M .* HrP
        else
            PY.=w_norm_A .* HrA./γ .+ w_norm_M .* HrP./β
        end

        PrmSig.=Prm.^(1-σ)
        PraSig.=Pra.^(1-σ)
        Get_T(L,c, Prms, Pras,tm, ta,PrrmSig,PrraSig,PrmSig, PraSig, PY,τs,σ)
        # Market clearing wages, set price index in region 1 as baseline
        wprime_A.= γ .* α     .* ta./HrA
        if β==0.0
            wprime_M.= (1-α) .* tm./HrP
        else
            wprime_M.= β .* (1-α) .* tm./HrP
        end

        # calculate tolerance
        diff_w =100*max(findmax(abs.(wprime_A.-w_norm_A)./w_norm_A)[1],
            findmax(abs.(wprime_M.-w_norm_M)./w_norm_M)[1])
                        #println(diff_w)

        # update wage guess
          wguess_A .=((1 .-convergence_speed) .* wprime_A.+ convergence_speed .*w_norm_A)
          wguess_M .=((1 .-convergence_speed) .* wprime_M.+ convergence_speed .*w_norm_M)

        # update iteration counter
            iter_w=iter_w+1
               if iter_w==iters-1
            println("Max iters exceeded--static clearing in iter_w")
        end
        end # this concludes the inner loop, gets w fixed point


                    wguess_A.=wguess_A./T_Ψguess[1]
                    wguess_M.=wguess_M./T_Ψguess[1]
            T_Ψguess.=T_Ψguess./T_Ψguess[1]
            diff_Ψ=100*findmax(abs.(T_Ψguess-Ψguess)./Ψguess)[1]
            Ψguess.=((1 .-convergence_psi).*T_Ψguess+convergence_psi.*Ψguess)
            iter_Ψ=iter_Ψ+1
        if iter_Ψ==iters-1
            println("Max iters exceeded--static clearing in iter_Ψ")
        end

    end # this concludes the main loop, gets Ψ fixed point
    mi.=mni.+mri
    mf.=mnf.+mrf

    S=wage_clearing_output9(wguess_A, wguess_M, LNI_p, LNF_p, LRI_p, LRF_p, Nactive, NoEntry, mi, mf,Ψguess,
                HrP, HrA)

    return S
end



#=
function Qmat_generator(shockpath, Pv::param, Q, Time)

    @unpack θ, χ, ϕ_fa, ϕ_fm, ϕ_ia, ϕ_im, ψ, ϵ, σ, α, ɷ, ϱ, ρ, γ, β, λ, δ, fE=Pv

    Qmat=zeros(length(Q), Time)
    Qmat[:,1].=log.(Q)
    for t in 2:Time
            Qmat[:,t] .= (1-ϱ).* log.(Q) .+ ϱ .*Qmat[:,t-1] .+ ɷ .*shockpath[:,t]
    end

    Qmat.= exp.(Qmat)
    return Qmat
end
=#


function Random_Sample(DF::DataFrame, Sample_Length)

    L=length(DF[:,1])
      if Sample_Length >L
        println("Sample size too big")
    end

    # thank you https://discourse.julialang.org/t/how-to-sample-rows-from-a-dataframe/19782
    DF2=DF[sample(MersenneTwister(1234),axes(DF, 1), Sample_Length; replace = false, ordered = true), :]
    return DF2
end


function SSD(x,y)  # Percentage squared difference
    D=((x-y)/y)^2
    return D
end

function ASD(x,y) # Absolute squared difference
    D=(x-y)^2
    return D
end


# New vector of "true moments": longer than before
True_Moments=(Ref_Wage_Ratio_log=-0.075,
            Ref_manu_gr_50=0.317,
            Ref_manu_gr_61=0.241,
            Ref_GDP_gr_50=-0.083,
            Ref_GDP_gr_61=0.502,
            Ref_pop_gr_50=1.359,
            Ref_pop_gr_61=1.029,
            Ref_sh_cor_50=0.735,
            Ref_sh_cor_61=0.586,
            GDP_lag_coef=0.555,
            GDP_lag_sd=0.041,
            Manuf_Ag_WageRatio=1.5,
            In_Out_Flow_Ratio=0.73,
            Mobility_Distance=0.67,
            Distance_coef=0.06,
            Ref_pop_gr_50_55=-0.342,
            Ref_pop_gr_55_61=-0.087,
            Ref_GDP_gr_80=0.201,
            pop_coef_80_UB=1.2,
            Ref_pop_gr_80=1.0
            )

